<?php

/**
 * i-doit
 *
 * API model.
 *
 * @package     i-doit
 * @subpackage  API
 * @author      Dennis Stücken <dstuecken@synetics.de>
 * @copyright   synetics GmbH
 * @license     http://www.i-doit.com/license
 */
class isys_api_model_cmdb_object extends isys_api_model_cmdb implements isys_api_model_interface
{
    /**
     * Data formatting used in format methods.
     *
     * @var  array
     */
    protected $m_mapping = [
        'isys_obj__id'                     => 'id',
        'isys_obj__title'                  => 'title',
        'isys_obj__sysid'                  => 'sysid',
        'isys_obj__isys_obj_type__id'      => 'objecttype',
        'isys_obj_type__title'             => ['_L', 'type_title'],
        'isys_obj_type__icon'              => 'type_icon',
        'isys_obj__status'                 => 'status',
        'isys_obj__isys_cmdb_status__id'   => 'cmdb_status',
        'isys_cmdb_status__title'          => ['_L', 'cmdb_status_title'],
        'isys_obj__created'                => 'created',
        'isys_obj__updated'                => 'updated',
        'isys_catg_image_list__image_link' => ['isys_api_model_cmdb_objects::get_image_url', 'image']
    ];

    /**
     * Possible options and their parameters.
     *
     * @var  array
     */
    protected $m_options = [
        'read'       => [
            'id' => [
                'type'        => 'int',
                'description' => 'Object id',
                'reference'   => 'isys_obj__id',
                'optional'    => false,
            ],
        ],
        'create'     => [
            'title'       => [
                'type'        => 'string',
                'description' => 'Object title',
                'reference'   => 'isys_obj__title',
                'optional'    => false,
            ],
            'type'        => [
                'type'        => 'int|string',
                'description' => 'Object type as string constant or id',
                'reference'   => 'isys_obj_type__id',
                'optional'    => false,
            ],
            'cmdb_status' => [
                'type'        => 'int|string',
                'description' => 'Cmdb status id or constant',
                'reference'   => 'isys_obj__isys_cmdb_status__id',
                'optional'    => true,
            ],
        ],
        'update'     => [
            'id' => [
                'type'        => 'int',
                'description' => 'Object id',
                'reference'   => 'isys_obj__id',
                'optional'    => false,
            ],
        ],
        'delete'     => [
            'id'     => [
                'type'        => 'int',
                'description' => 'Object id',
                'reference'   => 'isys_obj__id',
                'optional'    => false,
            ],
            'status' => [
                'type'        => 'int',
                'description' => 'Object status',
                'reference'   => 'isys_obj__status',
                'optional'    => true,
            ],
        ],
        'quickpurge' => [
            'id' => [
                'type'        => 'int',
                'description' => 'Object id',
                'reference'   => 'isys_obj__id',
                'optional'    => false,
            ],
        ],
    ];

    /**
     * Validation.
     *
     * @var  array
     */
    protected $m_validation = [];

    /**
     * Fetches information about an object.
     *
     * @param   array $p_params Parameters. Structure: array('id' => 1).
     *
     * @return  array  Returns an empty array when an error occures.
     */
    public function read($p_params)
    {
        $l_return = [];

        if (isset($p_params[C__CMDB__GET__OBJECT]))
        {
            $p_params['id'] = $p_params[C__CMDB__GET__OBJECT];
            unset($p_params[C__CMDB__GET__OBJECT]);
        }

        if (isset($p_params['id']) && $p_params['id'])
        {
            $this->m_log->info('Retrieving object with id ' . $p_params['id']);

            if ($this->useAuth)
            {
                isys_auth_cmdb::instance()->check_rights_obj_and_category(isys_auth::VIEW, $p_params['id'], 'C__CATG__GLOBAL');
            }

            // Data retrieval.
            $l_data = $this->m_dao->get_object_by_id($p_params['id']);

            // Data formatting.
            if ($l_data->count() > 0)
            {
                return $this->format_by_mapping($this->m_mapping, $l_data->get_row());
            }
        }
        else
        {
            $this->m_log->error('Object ID missing.');
        }

        return $l_return;
    }

    /**
     * Creates an object.
     *
     * @param   array $p_params Parameters (depends on data method).
     *
     * @return  array
     * @throws  isys_exception_api
     */
    public function create($p_params)
    {
        $l_return = [];

        if (!isset($p_params['title']))
        {
            throw new isys_exception_api('Object title missing');
        }

        if (! isset($p_params['type']) || empty($p_params['type']) || !$p_params['type'])
        {
            throw new isys_exception_api('Object type missing');
        }

        try
        {
            // Object-Type.
            $p_params['type'] = is_numeric($p_params['type']) ? $p_params['type'] : (defined($p_params['type']) ? constant($p_params['type']) : null);

            if ($this->useAuth)
            {
                isys_auth_cmdb_objects::instance()->is_object_type_allowed(isys_auth::CREATE, $p_params['type']);
            }

            // CMDB-Status.
            $p_params['cmdb_status'] = is_numeric($p_params['cmdb_status']) ?
                $p_params['cmdb_status'] :
                (defined($p_params['cmdb_status']) ? constant($p_params['cmdb_status']) : null);

            if ($p_params['type'] > 0 && $this->m_dao->get_objtype($p_params['type'])->num_rows() > 0)
            {
                // insert the object.
                $l_return['id'] = $this->m_dao->insert_new_obj(
                    $p_params['type'],
                    false,
                    $p_params['title'],
                    null,
                    C__RECORD_STATUS__NORMAL,
                    null,
                    null,
                    false,
                    null,
                    null,
                    null,
                    null,
                    $p_params['category'],
                    $p_params['purpose'],
                    $p_params['cmdb_status'],
                    $p_params['description']
                );

                if (isset($p_params['defaultTemplate']) && $p_params['defaultTemplate'])
                {
                    // Get template module.
                    $l_template_module  = new isys_module_templates();
                    $l_default_template = $this->m_dao->get_default_template_by_obj_type($p_params['type']);

                    $l_template_module->create_from_template(
                        [$l_default_template],
                        $p_params['type'],
                        $p_params['title'],
                        $l_return['id'],
                        false,
                        1,
                        '',
                        $p_params['category'],
                        $p_params['purpose']
                    );
                }

                if ($l_return['id'] > 0)
                {
                    // @see API-15 Force INTEGER type case.
                    $l_return['id'] = (int) $l_return['id'];

                    // Create logbook entry.
                    isys_event_manager::getInstance()
                        ->triggerCMDBEvent(
                            'C__LOGBOOK_EVENT__OBJECT_CREATED',
                            '-object initialized-',
                            $l_return['id'],
                            $this->m_dao->get_objTypeID($l_return['id'])
                        );

                    $l_return['message'] = 'Object was successfully created';

                    // @see  API-38  Index the newly created object.
                    if (class_exists('\\idoit\\Module\\Cmdb\\Search\\IndexExtension\\OnelineIndexer')) {
                        // @todo  Cleanup. This code is for i-doit < 1.11
                        (new \idoit\Module\Cmdb\Search\IndexExtension\OnelineIndexer($this->m_db))
                            ->setObjectIds([$l_return['id']])
                            ->index();
                    } else if (class_exists('\\idoit\\Module\\Cmdb\\Search\\Index\\Data\\CategoryCollector')) {
                        try {
                            // @todo  Cleanup. This code is for i-doit >= 1.11
                            $collector = new idoit\Module\Cmdb\Search\Index\Data\CategoryCollector($this->m_db, ['C__CATG__GLOBAL'], [$l_return['id']]);
                            $engine = new \idoit\Module\Search\Index\Engine\Mysql($this->m_db);
                            $manager = new \idoit\Module\Search\Index\Manager($engine, [$collector]);

                            $manager->setMode(\idoit\Module\Search\Index\Manager::MODE_CREATION);
                            $manager->generateIndex();
                        } catch (Exception $e) {
                            $l_return['message'] .= '. But the search index could not be created: ' . $e->getMessage();
                        }
                    }

                    $l_return['success'] = true;
                }
                else
                {
                    $l_return['message'] = 'Error while creating object';
                    $l_return['success'] = false;
                }
            }
            else
            {
                throw new isys_exception_api('Object type not found.');
            }
        }
        catch (isys_exception_cmdb $e)
        {
            throw new isys_exception_api($e->getMessage());
        }

        return $l_return;
    }

    /**
     * Deletes an object.
     *
     * @param   array $p_params Parameters (depends on data method).
     *
     * @return  isys_api_model_cmdb  Returns itself.
     * @throws  isys_exception_api
     */
    public function delete($p_params)
    {
        $l_return = [
            'message' => 'Error while deleting object(s)',
            'success' => false
        ];

        if (isset($p_params['ids']))
        {
            $p_params['id'] = $p_params['ids'];
        }

        if (!isset($p_params['id']))
        {
            throw new isys_exception_api('Object id missing');
        }
        else
        {
            if (is_numeric($p_params['id']))
            {
                $l_id = $p_params['id'];
                unset($p_params['id']);
                $p_params['id'] = [$l_id];
            }

            if (is_array($p_params['id']))
            {
                $l_status = C__RECORD_STATUS__DELETED;
                if (isset($p_params['status']))
                {
                    if (is_string($p_params['status']) && defined($p_params['status']))
                    {
                        $p_params['status'] = constant($p_params['status']);
                    }

                    if (($p_params['status'] == C__RECORD_STATUS__DELETED || $p_params['status'] == C__RECORD_STATUS__ARCHIVED ||
                        $p_params['status'] == C__RECORD_STATUS__PURGE))
                    {
                        $l_status = $p_params['status'];
                    }
                }

                switch ($l_status)
                {
                    case C__RECORD_STATUS__PURGE:
                        $message  = 'purged';
                        break;

                    case C__RECORD_STATUS__ARCHIVED:
                        $message  = 'archived';
                        break;

                    default:
                    case C__RECORD_STATUS__DELETED:
                        $message  = 'deleted';
                        break;
                }

                $skipped = [];

                foreach ($p_params['id'] as $l_id)
                {
                    if (!$this->m_dao->obj_exists($l_id))
                    {
                        $skipped[] = '#' . $l_id;

                        continue;
                    }

                    switch ($l_status)
                    {
                        case C__RECORD_STATUS__PURGE:
                            $l_logbookEvent = 'C__LOGBOOK_EVENT__OBJECT_PURGED';
                            $l_result       = $this->m_dao->delete_object($l_id);
                            break;
                        case C__RECORD_STATUS__ARCHIVED:
                            $l_logbookEvent = 'C__LOGBOOK_EVENT__OBJECT_ARCHIVED';
                            $l_result       = $this->m_dao->set_object_status($l_id, $l_status);
                            break;
                        default:
                        case C__RECORD_STATUS__DELETED:
                            $l_logbookEvent = 'C__LOGBOOK_EVENT__OBJECT_DELETED';
                            $l_result       = $this->m_dao->set_object_status($l_id, $l_status);
                            break;
                    }

                    if (!$l_result)
                    {
                        throw new isys_exception_api(sprintf('Error deleting object with id %s', $l_id));
                    }

                    isys_event_manager::getInstance()
                        ->triggerCMDBEvent(
                            $l_logbookEvent,
                            $this->m_dao->get_last_query(),
                            $l_id,
                            $this->m_dao->get_objTypeID($l_id)
                        );
                }

                $l_return['message'] = 'Object(s) successfully ' . $message . '!' . (count($skipped) ? ' But skipped ' . implode(', ', $skipped)  . '.': '');
                $l_return['success'] = true;
            }
        }

        return $l_return;
    }

    /**
     * Updates data.
     *
     * @param   array $p_params Parameters (depends on data method).
     *
     * @return  isys_api_model_cmdb  Returns itself.
     * @throws  isys_exception_api
     */
    public function update($p_params)
    {
        $l_return = [];

        if (!isset($p_params['id']))
        {
            throw new isys_exception_api('Object id missing');
        }
        else
        {
            $l_dao_global = new isys_cmdb_dao_category_g_global($this->m_dao->get_database_component());

            if ($l_dao_global->save_title($p_params['id'], $p_params['title']))
            {
                // Create logbook entry.
                // @todo type should be retrieved by object ID.
                isys_event_manager::getInstance()
                    ->triggerCMDBEvent(
                        'C__LOGBOOK_EVENT__CATEGORY_CHANGED',
                        '- Object title changed to ' . $p_params['title'] . ' -',
                        $p_params['id'],
                        $this->m_dao->get_objTypeID($p_params['id']),
                        'LC__CMDB__CATG__GLOBAL'
                    );

                $l_dao_global->object_changed($p_params['id']);
                $l_return['message'] = 'Object title was successfully updated';
                $l_return['success'] = true;

                return $l_return;
            }
        }

        $l_return['message'] = 'Error while updating object';
        $l_return['success'] = false;

        return $l_return;
    }

    /**
     * Deletes an object.
     *
     * @param   array $p_params Parameters (depends on data method).
     *
     * @return  isys_api_model_cmdb  Returns itself.
     * @throws  isys_exception_api
     */
    public function quickpurge($p_params)
    {
        if (!isys_settings::get('cmdb.quickpurge', false))
        {
            throw new isys_exception_api('Quickpurge is not enabled');
        }

        $l_return = [
            'message' => 'Error while quickpurging object(s)',
            'success' => false,
        ];

        if (isset($p_params['ids']))
        {
            $p_params['id'] = $p_params['ids'];
        }

        if (!isset($p_params['id']))
        {
            throw new isys_exception_api('Object id missing');
        }
        else
        {
            if (is_numeric($p_params['id']))
            {
                $l_id = $p_params['id'];
                unset($p_params['id']);
                $p_params['id'] = [$l_id];
            }

            if (is_array($p_params['id']))
            {
                $skipped = [];

                foreach ($p_params['id'] as $l_id)
                {
                    if (!$this->m_dao->obj_exists($l_id))
                    {
                        $skipped[] = '#' . $l_id;

                        continue;
                    }

                    if (!$this->m_dao->rank_record($l_id, C__CMDB__RANK__DIRECTION_DELETE, 'isys_obj', null, true))
                    {
                        throw new isys_exception_api(sprintf('Error while purging object with id %s', $l_id));
                    }
                }

                $l_return['message'] = 'Object(s) successfully purged!' . (count($skipped) ? ' But skipped ' . implode(', ', $skipped) : '') . '.';
                $l_return['success'] = true;
            }
        }

        return $l_return;
    }

    /**
     * Constructor.
     *
     * @param  isys_cmdb_dao $p_dao
     */
    public function __construct(isys_cmdb_dao &$p_dao)
    {
        $this->m_dao = $p_dao;
        parent::__construct();
    }
}